/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Class
    dynSmagorinsky

Description
<pre>
    The Isochoric dynamic Smagorinsky Model
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Algebraic eddy viscosity SGS model founded on the assumption that
    local equilibrium prevails, hence

        B = 2/3*k*I - 2*nuEff*dev(D)

    where

       k = cI*delta^2*||D||^2
       nuEff = cD*delta^2*||D|| + nu

    In the dynamic version of the choric  Smagorinsky model
    the coefficients cI and cD are calculated during the simulation,

        cI=<K*m>/<m*m>

    and

        cD=<L.M>/<M.M>,

    where

        K = 0.5*(F(U.U) - F(U).F(U))
        m = delta^2*(4*||F(D)||^2 - F(||D||^2))
        L = rho*(F(U*U) - F(U)*F(U) - 0.33*K*I)
        M = rho*delta^2*(F(||D||*dev(D)) - 4*||F(D)||*F(dev(D)))
</pre>

SourceFiles
    dynSmagorinsky.C

\*---------------------------------------------------------------------------*/

#ifndef dynSmagorinsky_H
#define dynSmagorinsky_H

#include "Smagorinsky.H"
#include "LESfilter.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace LESmodels
{

/*---------------------------------------------------------------------------*\
                           Class dynSmagorinsky Declaration
\*---------------------------------------------------------------------------*/

class dynSmagorinsky
:
    public GenEddyVisc
{
    // Private data

        volScalarField k_;

        autoPtr<LESfilter> filterPtr_;
        LESfilter& filter_;


    // Private Member Functions

        //- Calculate coefficients cD, cI from filtering velocity field
        dimensionedScalar cD(const volTensorField& D) const;
        dimensionedScalar cI(const volTensorField& D) const;

        // Disallow default bitwise copy construct and assignment
        dynSmagorinsky(const dynSmagorinsky&);
        dynSmagorinsky& operator=(const dynSmagorinsky&);


public:

    //- Runtime type information
    TypeName("dynSmagorinsky");

    // Constructors

        //- Constructor from components
        dynSmagorinsky
        (
            const volVectorField& U,
            const surfaceScalarField& phi,
            transportModel& transport
        );


    // Destructor

        ~dynSmagorinsky();


    // Member Functions

        //- Return SGS kinetic energy
        tmp<volScalarField> k() const
        {
            return k_;
        }

        //- Correct Eddy-Viscosity and related properties
        void correct(const tmp<volTensorField>& gradU);

        //- Read turbulenceProperties dictionary
        bool read();
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace LESmodels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
