/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description
    A PtrList<class T> is a 1D array of pointers to objects
    of T 'T', where the size of the array is known and used for
    subscript bounds checking, etc.
    The element operator [] returns a reference to the object
    rather than a pointer

\*---------------------------------------------------------------------------*/

#include "error.H"

#include "PtrList.H"
#include "PtrListLoopM.H"
#include "SLPtrList.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * * //

// Construct with zero length
template<class T>
PtrList<T>::PtrList()
:
    ptrs_(),
    nextFree_(0)
{}


// Construct with length specified
template<class T>
PtrList<T>::PtrList(const label s)
:
    ptrs_(s, reinterpret_cast<T*>(NULL)),
    nextFree_(0)
{}


// Construct as copy
template<class T>
PtrList<T>::PtrList(const PtrList<T>& a)
:
    ptrs_(a.size()),
    nextFree_(a.nextFree_)
{
    forAll(*this, i)
    {
        ptrs_[i] = (a[i]).clone().ptr();
    }
}


// Construct as copy given additional information
template<class T>
template<class CloneArg>
PtrList<T>::PtrList(const PtrList<T>& a, const CloneArg& cloneArg)
:
    ptrs_(a.size()),
    nextFree_(a.nextFree_)
{
    forAll(*this, i)
    {
        ptrs_[i] = (a[i]).clone(cloneArg).ptr();
    }
}


// Construct as copy or re-use as specified.
template<class T>
PtrList<T>::PtrList(PtrList<T>& a, bool reUse)
:
    ptrs_(a.size()),
    nextFree_(a.nextFree_)
{
    if (reUse)
    {
        forAll(*this, i)
        {
            ptrs_[i] = a.ptrs_[i];
            a.ptrs_[i] = NULL;
        }
        a.nextFree_ = 0;
        a.setSize(0);
    }
    else
    {
        forAll(*this, i)
        {
            ptrs_[i] = (a[i]).clone().ptr();
        }
    }
}

// Construct as copy of SLPtrList<T>
template<class T>
PtrList<T>::PtrList(const SLPtrList<T>& sll)
:
    ptrs_(sll.size()),
    nextFree_(sll.size())
{
    if (sll.size())
    {
        label i = 0;
        for
        (
            typename SLPtrList<T>::const_iterator iter = sll.begin();
            iter != sll.end();
            ++iter
        )
        {
            ptrs_[i++] = (iter()).clone().ptr();
        }
    }
}


// * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * * //

// Destroy ptrlist elements
template<class T>
PtrList<T>::~PtrList()
{
    forAll(*this, i)
    {
        if (ptrs_[i])
        {
            delete ptrs_[i];
        }
    }
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

// Return an element pointer to be set.  The pointer is checked
// if already set and the index range checked.
template<class T>
typename PtrList<T>::Tptr& PtrList<T>::set(const label i)
{
    if (ptrs_[i])
    {
        FatalErrorIn("PtrList::set(const label)")
            << "pointer already set, cannot set to new element"
            << abort(FatalError);
    }

    return ptrs_[i];
}


// Hook an an element created by new onto the list
template<class T>
void PtrList<T>::hook(T* ptr)
{
    if (nextFree_ < size())
    {
        set(nextFree_) = ptr;
    }
    else
    {
        FatalErrorIn("PtrList::hook(T*)")
            << "PtrList full, cannot add new element"
            << abort(FatalError);
    }

    nextFree_++;
}


template<class T>
void PtrList<T>::setSize(const label newSize)
{
    label oldSize = size();

    if (newSize == 0)
    {
        clear();
    }
    else if (newSize < oldSize)
    {
        register label i;
        for (i=newSize; i<oldSize; i++)
        {
            if (ptrs_[i])
            {
                delete ptrs_[i];
            }
        }

        ptrs_.setSize(newSize);
        nextFree_ = newSize;
    }
    else if (newSize > oldSize)
    {
        ptrs_.setSize(newSize);

        register label i;
        for (i=oldSize; i<newSize; i++)
        {
            ptrs_[i] = NULL;
        }
    }
}


template<class T>
void PtrList<T>::clear()
{
    forAll(*this, i)
    {
        if (ptrs_[i])
        {
            delete ptrs_[i];
        }
    }

    ptrs_.clear();

    // Reset the next free element to zero
    nextFree_ = 0;
}


// Transfer the contents of the argument List into this List
// and anull the argument list
template<class T>
void PtrList<T>::transfer(PtrList<T>& a)
{
    ptrs_.transfer(a.ptrs_);
    a.nextFree_ = 0;
}


template<class T>
void PtrList<T>::reorder(const UList<label>& oldToNew)
{
    if (oldToNew.size() != size())
    {
        FatalErrorIn("PtrList<T>::reorder(const UList<label>&)")
            << "Size of map (" << oldToNew.size()
            << ") not equal to list size (" << size()
            << ")." << abort(FatalError);
    }

    List<T*> newPtrs_(ptrs_.size(), reinterpret_cast<T*>(NULL));

    forAll(*this, i)
    {
        label newI = oldToNew[i];

        if (newI < 0 || newI >= size())
        {
            FatalErrorIn("PtrList<T>::reorder(const UList<label>&)")
                << "Illegal index " << newI << nl
                << "Valid indices are 0.." << size()-1
                << abort(FatalError);
        }

        if (newPtrs_[newI])
        {
            FatalErrorIn("PtrList<T>::reorder(const UList<label>&)")
                << "reorder map is not unique; element " << newI
                << " already set." << abort(FatalError);
        }
        newPtrs_[newI] = ptrs_[i];
    }

    forAll(newPtrs_, i)
    {
        if (!newPtrs_[i])
        {
            FatalErrorIn("PtrList<T>::reorder(const UList<label>&)")
                << "Element " << i << " not set after reordering." << nl
                << abort(FatalError);
        }
    }

    ptrs_.transfer(newPtrs_);
}


// * * * * * * * * * * * * * * * Member Operators  * * * * * * * * * * * * * //

// Assignment
template<class T>
PtrList<T>& PtrList<T>::operator=(const PtrList<T>& a)
{
    if (this == &a)
    {
        FatalErrorIn("PtrList<T>::operator=(const PtrList<T>&)")
            << "attempted assignment to self"
            << abort(FatalError);
    }

    if (size() == 0)
    {
        setSize(a.size());
        nextFree_ = a.nextFree_;

        forAll(*this, i)
        {
            ptrs_[i] = (a[i]).clone().ptr();
        }
    }
    else if (a.size() == size())
    {
        forAll(*this, i)
        {
            (*this)[i] = a[i];
        }
    }
    else
    {
        FatalErrorIn("PtrList::operator=(const PtrList<T>&)")
            << "bad size: " << a.size()
            << abort(FatalError);
    }


    return *this;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#include "PtrListIO.C"

// ************************************************************************* //
