/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Class
    SubField

Description
    SubField is a Field obtained as a section of another Field.  Thus it is
    itself unallocated so that no storage is allocated or de-allocated
    during it's use.  To achieve this behaviour, SubField is derived from
    UnallocatedField rather than Field.

SourceFiles
    SubFieldI.H

\*---------------------------------------------------------------------------*/

#ifndef SubField_H
#define SubField_H

#include "Field.H"
#include "SubList.H"
#include "Field.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

/*---------------------------------------------------------------------------*\
                           Class SubField Declaration
\*---------------------------------------------------------------------------*/

template<class Type>
class SubField
:
    public refCount,
    public SubList<Type>
{

public:

    //- Component type
    typedef typename pTraits<Type>::cmptType cmptType;


    // Constructors

        //- Construct from UnallocatedField and SubField size
        inline SubField
        (
            const SubList<Type>& slist
        );

        //- Construct from UnallocatedField and SubField size
        inline SubField
        (
            const UList<Type>& list,
            const label subSize
        );

        //- Construct from UnallocatedField, start and end indices
        inline SubField
        (
            const UList<Type>& list,
            const label subSize,
            const label startIndex
        );

        //- Construct as copy
        inline SubField
        (
            const SubField<Type>& sfield
        );


    // Member functions

        //- Return a null SubField
        static inline SubField<Type>& null();

        //- Return a component field of the field
        inline tmp<Field<cmptType> > component(const direction) const;

        //- Return the field transpose (only defined for second rank tensors)
        tmp<Field<Type> > T() const;


    // Member operators

        //- Assignment from UList operator. Takes linear time.
        inline void operator=(const SubField<Type>&);

        //- Allow cast to a const Field<Type>&
        inline operator const Field<Type>&() const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#include "SubFieldI.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
