/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Class
    ListOps

Description
    Various functions to operate on Lists.

SourceFiles
    ListOps.C
    ListOpsTemplates.C

\*---------------------------------------------------------------------------*/

#ifndef ListOps_H
#define ListOps_H

#include "labelList.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

//- Renumber the values (not the indices) of a list.
template<class List>
List renumber(const labelList& oldToNew, const List&);

//- Inplace renumber the values of a list.
template<class List>
void inplaceRenumber(const labelList& oldToNew, List&);


//- Reorder the elements (indices, not values) of a list. Does not do any
//  check for illegal indices (e.g. -1)
template<class List>
List reorder(const labelList& oldToNew, const List&);

//- Inplace reorder the elements of a list. Does not do any
//  check for illegal indices.
template<class List>
void inplaceReorder(const labelList& oldToNew, List&);


//- Extract elements of List whose region is certain value. Use e.g.
//  to extract all selected elements: 
//  extract<boolList, labelList>(selectedElems, true, lst);
template<class T, class List>
List extract(const UList<T>& regions, const T& region, const List&);


//- Invert one-to-one map. Unmapped elements will be -1.
labelList invert(const label len, const labelList& oldToNew);

//- Invert one-to-many map. Unmapped elements will be size 0.
labelListList invertOneToMany(const label len, const labelList&);

//- Create identity map (map[i] == i) of given length
labelList identity(const label len);

//- Find first occurence of given element and return index,
//  return -1 if not found. Linear search.
template<class List>
label findIndex(const List&, typename List::const_reference);


//- Find index of max element (and larger than given element).
//  return -1 if not found. Linear search.
template<class List>
label findMax(const List&);


//- Find index of min element (and less than given element).
//  return -1 if not found. Linear search.
template<class List>
label findMin(const List&);


//- Find first occurence of given element in sorted list and return index,
//  return -1 if not found. Binary search.
template<class List>
label findSortedIndex(const List&, typename List::const_reference);


//- Find last element < given value in sorted list and return index,
//  return -1 if not found. Binary search.
template<class List>
label findLower(const List&, typename List::const_reference);


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#ifdef NoRepository
#   include "ListOpsTemplates.C"
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //

