/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "Matrix.H"
#include "Istream.H"
#include "Ostream.H"
#include "token.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * * * Ostream Operator *  * * * * * * * * * * * * //

// Construct from Istream
template<class T>
Matrix<T>::Matrix(Istream& is)
:
    n_(0),
    m_(0),
    v_(0)
{
    operator>>(is, *this);
}


template<class T>
Istream& operator>>(Istream& is, Matrix<T>& M)
{
    // Anull matrix
    M.clear();

    is.fatalCheck("operator>>(Istream&, Matrix<T>&)");

    token firstToken(is);

    is.fatalCheck("operator>>(Istream&, Matrix<T>&) : reading first token");

    if (firstToken.isLabel())
    {
        M.n_ = firstToken.labelToken();
        M.m_ = readLabel(is);

        M.allocate();
        T* v = M.v_[0];

        // Read list contents depending on data format
        if (is.format() == IOstream::ASCII || !writeBinary(v))
        {
            // Read beginning of contents
            char listDelimiter = is.readBeginList("Matrix");

            label nm = M.n_*M.m_;

            if (nm)
            {
                if (listDelimiter == token::BEGIN_LIST)
                {
                    label k = 0;

                    // loop over rows
                    for (register label i=0; i< M.n(); i++)
                    {
                        listDelimiter = is.readBeginList("MatrixRow");

                        for (register label j=0; j< M.m(); j++)
                        {
                            is >> v[k++];

                            is.fatalCheck
                            (
                                "operator>>(Istream&, Matrix<T>&) : "
                                "reading entry"
                            );
                        }

                        is.readEndList("MatrixRow");
                    }
                }
                else
                {
                    T element;
                    is >> element;

                    is.fatalCheck
                    (
                        "operator>>(Istream&, Matrix<T>&) : "
                        "reading the single entry"
                    );

                    for (register label i=0; i<nm; i++)
                    {
                        v[i] = element;
                    }
                }
            }

            // Read end of contents
            is.readEndList("Matrix");
        }
        else
        {
            is.read(reinterpret_cast<char*>(v), M.n_*M.m_*sizeof(T));

            is.fatalCheck
            (
                "operator>>(Istream&, Matrix<T>&) : reading the binary block"
            );
        }
    }
    else
    {
        FatalIOErrorIn("operator>>(Istream&, Matrix<T>&)", is)
            << "incorrect first token, expected <int>, found "
            << firstToken.info()
            << exit(FatalIOError);
    }

    return is;
}


template<class T>
Ostream& operator<<(Ostream& os, const Matrix<T>& M)
{
    // Write list contents depending on data format
    if (os.format() == IOstream::ASCII || !writeBinary(M.v_))
    {
        bool uniform = false;

        label nm = M.n_*M.m_;
        const T* v = M.v_[0];

        if (nm > 1 && writeBinary(v))
        {
            uniform = true;

            for (register label i=0; i< nm; i++)
            {
                if (v[i] != v[0])
                {
                    uniform = false;
                    break;
                }
            }
        }

        if (uniform)
        {
            // Write size of list and start contents delimiter
            os  << M.n() << token::SPACE << M.m() << token::SPACE 
                << token::BEGIN_BLOCK;

            // Write list contents
            os << v[0];

            // Write end of contents delimiter
            os << token::END_BLOCK;
        }
        else if (nm < 10 && writeBinary(v))
        {
            // Write size of list and start contents delimiter
            os  << M.n() << token::SPACE << M.m() << token::SPACE 
                << token::BEGIN_LIST;

            label k = 0;

            // loop over rows
            for (register label i=0; i< M.n(); i++)
            {
                os  << token::BEGIN_LIST;

                // Write row
                for (register label j=0; j< M.m(); j++)
                {
                    if (j > 0) os << token::SPACE;
                    os << v[k++];
                }

                os << token::END_LIST;
            }

            // Write end of contents delimiter
            os << token::END_LIST;
        }
        else
        {
            // Write size of list and start contents delimiter
            os  << nl << M.n() << token::SPACE << M.m() << nl
                << token::BEGIN_LIST;

            label k = 0;

            // loop over rows
            for (register label i=0; i< M.n(); i++)
            {
                os  << nl << token::BEGIN_LIST;

                // Write row
                for (register label j=0; j< M.m(); j++)
                {
                    os << nl << v[k++];
                }

                os << nl << token::END_LIST;
            }

            // Write end of contents delimiter
            os << nl << token::END_LIST << nl;
        }
    }
    else
    {
        os << nl << M.n_ << token::SPACE << M.m_ << nl;
        os.write(reinterpret_cast<const char*>(M.v_[0]), M.n_*M.m_*sizeof(T));
    }

    // Check state of IOstream
    os.check("Ostream& operator<<(Ostream&, const Matrix&)");

    return os;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
