/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "face.H"
#include "triPointRef.H"
#include "mathematicalConstants.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

// Construct list of edge vectors
tmp<vectorField> face::calcEdges(const pointField& points) const
{
    tmp<vectorField> tedges(new vectorField(size()));
    vectorField& edges = tedges();

    forAll(*this, i)
    {
        label ni = fcIndex(i);

        point thisPt = points[operator[](i)];
        point nextPt = points[operator[](ni)];

        vector vec(nextPt - thisPt);
        vec /= Foam::mag(vec) + VSMALL;

        edges[i] = vec;
    }

    return tedges;
}


// Calculate cosine of angle between edges at point index
scalar face::edgeCos
(
    const vectorField& edges,
    const label index
) const
{
    label leftEdgeI = left(index);
    label rightEdgeI = right(index);

    // note negate on left edge to get correct left-pointing edge.
    return -(edges[leftEdgeI] & edges[rightEdgeI]);
}


// Find index of most concave angle
label face::mostConcaveAngle
(
    const pointField& points,
    const vectorField& edges,
    scalar& maxAngle
) const
{
    vector n(normal(points));

    label index = 0;
    maxAngle = -GREAT;

    forAll(edges, i)
    {
        label leftEdgeI = left(i);
        label rightEdgeI = right(i);

        vector edgeNormal = edges[rightEdgeI] ^ edges[leftEdgeI];

        scalar edgeCos = edges[leftEdgeI] & edges[rightEdgeI];
        scalar edgeAngle = acos(max(-1.0, min(1.0, edgeCos)));

        scalar angle;

        if ((edgeNormal & n) > 0)
        {
            // Concave angle.
            angle = mathematicalConstant::pi + edgeAngle;
        }
        else
        {
            // Convex angle. Note '-' to take into account that rightEdge
            // and leftEdge are head-to-tail connected.
            angle = mathematicalConstant::pi - edgeAngle;
        }

        if (angle > maxAngle)
        {
            maxAngle = angle;
            index = i;
        }
    }

    return index;
}


// Counts/Splits face in triangles/triangles+quads.
// Recurses over anything more complex than triangle or quad.
void face::split
(
    const face::splitMode mode,
    const pointField& points,
    label& triI,
    label& quadI,
    faceList& triFaces,
    faceList& quadFaces
) const
{
    if (size() <= 2)
    {
        FatalErrorIn
        (
            "face::split"
            "(const face::splitMode, const pointField&, label&, label&"
            ", faceList&, faceList&)"
        )
            << "Serious problem: asked to split a face with < 3 vertices"
            << abort(FatalError);
    }
    if (size() == 3)
    {
        // Triangle. Just copy.
        if ((mode == COUNTQUAD) || (mode == COUNTTRIANGLE))
        {
            triI++;
        }
        else
        {
            triFaces[triI++] = *this;
        }
    }
    else if (size() == 4)
    {
        if (mode == COUNTTRIANGLE)
        {
            triI += 2;
        }
        if (mode == COUNTQUAD)
        {
            quadI++;
        }
        else if (mode == SPLITTRIANGLE)
        {
            //  Start at point with largest internal angle.
            const vectorField edges(calcEdges(points));

            scalar minAngle;
            label startIndex = mostConcaveAngle(points, edges, minAngle);
  
            label nextIndex = fcIndex(startIndex);
            label splitIndex = fcIndex(nextIndex);

            // Create triangles
            face triFace(3);
            triFace[0] = operator[](startIndex);
            triFace[1] = operator[](nextIndex);
            triFace[2] = operator[](splitIndex);

            triFaces[triI++] = triFace;

            triFace[0] = operator[](splitIndex);
            triFace[1] = operator[](fcIndex(splitIndex));
            triFace[2] = operator[](startIndex);

            triFaces[triI++] = triFace;
        }
        else
        {
            quadFaces[quadI++] = *this;
        }
    }
    else
    {
        // General case. Like quad: search for largest internal angle.

        const vectorField edges(calcEdges(points));

        scalar minAngle = 1;
        label startIndex = mostConcaveAngle(points, edges, minAngle);

        scalar bisectAngle = minAngle/2;
        vector rightEdge = edges[right(startIndex)];

        //
        // Look for opposite point which as close as possible bisects angle
        //

        // split candidate starts two points away.
        label index = fcIndex(fcIndex(startIndex));

        label minIndex = index;
        scalar minDiff = Foam::mathematicalConstant::pi;

        for(label i = 0; i < size() - 3; i++)
        {
            vector splitEdge
            (
                points[operator[](index)]
              - points[operator[](startIndex)]
            );
            splitEdge /= Foam::mag(splitEdge) + VSMALL;

            const scalar splitCos = splitEdge & rightEdge;
            const scalar splitAngle = acos(max(-1.0, min(1.0, splitCos)));
            const scalar angleDiff = fabs(splitAngle - bisectAngle);
            
            if (angleDiff < minDiff)
            {
                minDiff = angleDiff;
                minIndex = index;
            }

            // go to next candidate
            index = fcIndex(index);
        }


        // Split into two subshapes.
        //     face1: startIndex to minIndex
        //     face2: minIndex to startIndex

        // Get sizes of the two subshapes
        label diff = 0;
        if (minIndex > startIndex)
        {
            diff = minIndex - startIndex;
        }
        else
        {
            // folded round
            diff = minIndex + size() - startIndex;
        }

        label nPoints1 = diff + 1;
        label nPoints2 = size() - diff + 1;

        // Collect face1 points
        face face1(nPoints1);

        index = startIndex;
        for (label i = 0; i < nPoints1; i++)
        {
            face1[i] = operator[](index);
            index = fcIndex(index);
        }

        // Collect face2 points
        face face2(nPoints2);

        index = minIndex;
        for (label i = 0; i < nPoints2; i++)
        {
            face2[i] = operator[](index);
            index = fcIndex(index);
        }

        // Split faces
        face1.split(mode, points, triI, quadI, triFaces, quadFaces);
        face2.split(mode, points, triI, quadI, triFaces, quadFaces);
    }
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void face::collapse()
{
    label ci = 0;
    for (label i=1; i<size(); i++)
    {
        if (operator[](i) != operator[](ci))
        {
            operator[](++ci) = operator[](i);
        }
    }

    if (operator[](ci) != operator[](0))
    {
        ci++;
    }

    setSize(ci);
}


point face::centre(const pointField& meshPoints) const
{
    // Calculate the centre by breaking the face into triangles and
    // area-weighted averaging their centres
    // Changed to deal with small concavity by using a central decomposition
    // 

    // If the face is a triangle, do a direct calculation to avoid round-off
    // error-related problems
    // 
    if (size() == 3)
    {
        return triPointRef
        (
            meshPoints[operator[](0)],
            meshPoints[operator[](1)],
            meshPoints[operator[](2)]
        ).centre();
    }

    scalar sumA = 0;
    vector sumAc = vector::zero;

    point centrePoint = average(points(meshPoints));

    label nPoints = size();

    point nextPoint = centrePoint;

    register label pI;

    for (pI = 0; pI < nPoints; pI++)
    {
        if (pI < nPoints - 1)
        {
            nextPoint = meshPoints[operator[](pI + 1)];
        }
        else
        {
            nextPoint = meshPoints[operator[](0)];
        }

        // Note: for best accuracy, centre point always comes last
        // 
        vector n = triPointRef
        (
            meshPoints[operator[](pI)],
            nextPoint,
            centrePoint
        ).centre();

        scalar a = triPointRef
        (
            meshPoints[operator[](pI)],
            nextPoint,
            centrePoint
        ).mag();

        sumA += a;
        sumAc += a*n;
    }

    return sumAc/(sumA + VSMALL);
}


vector face::normal(const pointField& meshPoints) const
{
    // Calculate the normal by summing the face triangle normals.
    // Changed to deal with small concavity by using a central decomposition
    // 

    // If the face is a triangle, do a direct calculation to avoid round-off
    // error-related problems
    // 
    if (size() == 3)
    {
        return triPointRef
        (
            meshPoints[operator[](0)],
            meshPoints[operator[](1)],
            meshPoints[operator[](2)]
        ).normal();
    }

    vector n = vector::zero;

    point centrePoint = average(points(meshPoints));

    label nPoints = size();

    point nextPoint = centrePoint;

    register label pI;

    for (pI = 0; pI < nPoints; pI++)
    {
        if (pI < nPoints - 1)
        {
            nextPoint = meshPoints[operator[](pI + 1)];
        }
        else
        {
            nextPoint = meshPoints[operator[](0)];
        }

        // Note: for best accuracy, centre point always comes last
        // 
        n += triPointRef
        (
            meshPoints[operator[](pI)],
            nextPoint,
            centrePoint
        ).normal();
    }

    return n;
}


face face::reverseFace() const
{
    // reverse the label list and return
    // Changed to make sure that the starting point of the original
    // and the reverse face is identical.
    // 

    const labelList& myList = *this;

    labelList newList(size());

    newList[0] = myList[0];

    for (label pointI = 1; pointI < newList.size(); pointI++)
    {
        newList[pointI] = myList[size() - pointI];
    }

    return face(newList);
}


label face::which(const label globalIndex) const
{
    label pointInFace = -1;
    const labelList& f = *this;

    forAll (f, i)
    {
        if (f[i] == globalIndex)
        {
            pointInFace = i;
            break;
        }
    }

    return pointInFace;
}


scalar face::sweptVol
(
    const pointField& oldPoints,
    const pointField& newPoints
) const
{
    scalar sv = 0;

    // Calculate the swept volume by breaking the face into triangles and
    // summing their swept volumes.
    // Changed to deal with small concavity by using a central decomposition
    // 

    point centreOldPoint = average(points(oldPoints));

    point centreNewPoint = average(points(newPoints));

    label nPoints = size();

    point nextOldPoint = centreOldPoint;

    point nextNewPoint = centreNewPoint;

    register label pI;

    for (pI = 0; pI < nPoints; pI++)
    {
        if (pI < nPoints - 1)
        {
            nextOldPoint = oldPoints[operator[](pI + 1)];
            nextNewPoint = newPoints[operator[](pI + 1)];
        }
        else
        {
            nextOldPoint = oldPoints[operator[](0)];
            nextNewPoint = newPoints[operator[](0)];
        }

        // Note: for best accuracy, centre point always comes last
        // 
        sv += triPointRef
              (
                  oldPoints[operator[](pI)],
                  nextOldPoint,
                  centreOldPoint
              ).sweptVol
              (
                  triPointRef
                  (
                      newPoints[operator[](pI)],
                      nextNewPoint,
                      centreNewPoint
                  )
              );
    }

    return sv;
}


edgeList face::edges() const
{
    const labelList& points = *this;

    edgeList e(points.size());

    label pointI;

    for (pointI = 0; pointI < points.size() - 1; pointI++)
    {
        e[pointI] = edge(points[pointI], points[pointI + 1]);
    }

    // add last edge
    e[points.size() - 1] = edge(points[points.size() - 1], points[0]);

    return e;
}


// Number of triangles directly known from number of vertices
label face::nTriangles
(
    const pointField&
) const
{
    return size() - 2;
}


void face::triangles
(
    const pointField& points,
    label& triI,
    faceList& triFaces
) const
{
    faceList quadFaces;
    label quadI = 0;

    split(SPLITTRIANGLE, points, triI, quadI, triFaces, quadFaces);
}


void face::nTrianglesQuads
(
    const pointField& points,
    label& triI,
    label& quadI
) const
{
    faceList triFaces;
    faceList quadFaces;

    split(COUNTQUAD, points, triI, quadI, triFaces, quadFaces);
}


void face::trianglesQuads
(
    const pointField& points,
    label& triI,
    label& quadI,
    faceList& triFaces,
    faceList& quadFaces
) const
{
    split(SPLITQUAD, points, triI, quadI, triFaces, quadFaces);
}


// * * * * * * * * * * * * * * * Friend Operators  * * * * * * * * * * * * * //

bool operator==(const face& a, const face& b)
{
    // Basic rule: we assume that the sequence of labels in each list
    // will be circular in the same order (but not necessarily in the
    // same direction or from the same starting point).

    // Trivial reject: faces are different size
    label sizeA = a.size();
    label sizeB = b.size();

    if (sizeA != sizeB)
    {
        return false;
    }


    // Full list comparison
    const label firstA = a[0];
    label Bptr = -1;

    forAll (b, i)
    {
        if (b[i] == firstA)
        {
            Bptr = i;        // 'found match' at element 'i'
            break;
        }
    }

    // If no match was found, exit false
    if (Bptr < 0)
    {
        return false;
    }

    // Now we must look for the direction, if any
    label secondA = a[1];

    if (secondA == firstA || firstA == a[sizeA - 1])
    {
        face ca = a;
        ca.collapse();

        face cb = b;
        cb.collapse();

        return operator==(ca, cb);
    }

    label dir = 0;

    // Check whether at top of list
    Bptr++;
    if (Bptr == b.size())
    {
        Bptr = 0;
    }

    // Test whether upward label matches second A label
    if (b[Bptr] == secondA)
    {
        // Yes - direction is 'up'
        dir = 1;
    }
    else
    {
        // No - so look downwards, checking whether at bottom of list
        Bptr -= 2;
        if (Bptr < 0)
        {
            // Case (1) Bptr=-1
            if (Bptr == -1)
            {
                Bptr = b.size() - 1;
            }

            // Case (2) Bptr = -2
            else
            {
                Bptr = b.size() - 2;
            }
        }

        // Test whether downward label matches second A label
        if (b[Bptr] == secondA)
        {
            // Yes - direction is 'down'
            dir = -1;
        }
    }

    // Check whether a match was made at all, and exit false if not
    if (dir == 0)
    {
        return false;
    }

    // Decrement size by 2 to account for first searches
    sizeA -= 2;

    // We now have both direction of search and next element
    // to search, so we can continue search until no more points.
    label Aptr = 1;
    if (dir > 0)
    {
        while (sizeA--)
        {
            Aptr++;
            if (Aptr >= a.size())
            {
                Aptr = 0;
            }

            Bptr++;
            if (Bptr >= b.size())
            {
                Bptr = 0;
            }
            if (a[Aptr] != b[Bptr])
            {
                return false;
            }
        }
    }
    else
    {
        while (sizeA--)
        {
            Aptr++;
            if (Aptr >= a.size())
            {
                Aptr = 0;
            }

            Bptr--;
            if (Bptr < 0)
            {
                Bptr = b.size() - 1;
            }
            if (a[Aptr] != b[Bptr])
            {
                return false;
            }
        }
    }

    // They must be equal
    return true;
}


bool operator!=(const face& a, const face& b)
{
    return (!(a == b));
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
