/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "primitiveMesh.H"
#include "DynamicList.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

void primitiveMesh::calcCellEdges() const
{
    // Loop through all faces and mark up cells with edges of the face.
    // Check for duplicates

    if (debug)
    {
        Pout<< "primitiveMesh::calcCellEdges() : "
            << "calculating cellEdges"
            << endl;
    }

    // It is an error to attempt to recalculate cellEdges
    // if the pointer is already set
    if (cePtr_)
    {
        FatalErrorIn("primitiveMesh::calcCellEdges() const")
            << "cellEdges already calculated"
            << abort(FatalError);
    }
    else
    {
        // Set up temporary storage
        List<DynamicList<label, edgesPerCell_> > ce(nCells());


        // Get reference to faceCells and faceEdges
        const labelList& own = faceOwner();
        const labelList& nei = faceNeighbour();
        const labelListList& fe = faceEdges();

        // loop through the list again and add edges; checking for duplicates
        forAll (own, faceI)
        {
            const labelList& curEdges = fe[faceI];

            forAll (curEdges, edgeI)
            {
                bool found = false;

                DynamicList<label, edgesPerCell_>& curCellEdges =
                    ce[own[faceI]];

                forAll (curCellEdges, edgeOfCellI)
                {
                    if (curCellEdges[edgeOfCellI] == curEdges[edgeI])
                    {
                        found = true;
                        break;
                    }
                }

                if (!found)
                {
                    // Add the edge
                    curCellEdges.append(curEdges[edgeI]);
                }
            }
        }

        forAll (nei, faceI)
        {
            const labelList& curEdges = fe[faceI];

            forAll (curEdges, edgeI)
            {
                bool found = false;

                DynamicList<label, edgesPerCell_>& curCellEdges =
                    ce[nei[faceI]];

                forAll (curCellEdges, edgeOfCellI)
                {
                    if (curCellEdges[edgeOfCellI] == curEdges[edgeI])
                    {
                        found = true;
                        break;
                    }
                }

                if (!found)
                {
                    // add the edge
                    curCellEdges.append(curEdges[edgeI]);
                }
            }
        }

        cePtr_ = new labelListList(ce.size());
        labelListList& cellEdgeAddr = *cePtr_;

        // reset the size
        forAll (ce, cellI)
        {
            cellEdgeAddr[cellI].transfer(ce[cellI].shrink());
        }
    }
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

const labelListList& primitiveMesh::cellEdges() const
{
    if (!cePtr_)
    {
        calcCellEdges();
    }

    return *cePtr_;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
