/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description
    Restriction, prolongation and scaling for the amg solver.

\*---------------------------------------------------------------------------*/

#include "amgSymSolver.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

// restrict the fine mesh field to coarse mesh
void amgSymSolver::restrictField
(
    scalarField& result,
    const scalarField& fineField,
    const label fineLevelIndex
) const
{
    // Restriction operator is zero-order (summation)

    // Get addressing
    const labelField& fineToCoarse = restrictAddressing_[fineLevelIndex];

    // Debugging
    if (fineField.size() != fineToCoarse.size())
    {
        FatalErrorIn
        (
            "scalarField amgSymSolver::restrictField(const scalarField& f, "
            "const label fineLevelIndex) const"
        )   << "field does not correspond to level " << fineLevelIndex
            << " sizes: field = " << fineField.size() << " level = "
            << fineToCoarse.size()
            << abort(FatalError);
    }

    result = 0.0;

    forAll (fineField, i)
    {
        result[fineToCoarse[i]] += fineField[i];
    }
}


void amgSymSolver::prolongField
(
    scalarField& result,
    const scalarField& f,
    const label coarseLevelIndex
) const
{
    // Prolongation operator is zero order

    // Get addressing
    const labelField& fineToCoarse = restrictAddressing_[coarseLevelIndex];

    forAll(fineToCoarse, i)
    {
        result[i] = f[fineToCoarse[i]];
    }
}


scalar amgSymSolver::scalingFactor
(
    const scalarField& field,
    const scalarField& source,
    const scalarField& A
) const
{
    scalar scalingFactorNum = 0.0;
    scalar scalingFactorDenom = 0.0;

    forAll(field, i)
    {
        scalingFactorNum += source[i]*field[i];
        scalingFactorDenom += A[i]*field[i];
    }

    vector scalingVector(scalingFactorNum, scalingFactorDenom, 0);
    reduce(scalingVector, sumOp<vector>());    
    return scalingVector[0]/stabilise(scalingVector[1], VSMALL);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
