/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "PrimitivePatch.H"
#include "Map.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

template<class Face, template<class> class FaceList, class PointField>
void PrimitivePatch<Face, FaceList, PointField>::calcMeshData() const
{
    if (debug)
    {
        Info<< "PrimitivePatch<Face, FaceList, PointField>::"
               "calcMeshData() : "
               "calculating mesh data in PrimitivePatch"
            << endl;
    }

    // It is considered an error to attempt to recalculate meshPoints
    // if they have already been calculated.
    if (meshPointsPtr_ || localFacesPtr_)
    {
        FatalErrorIn
        (
            "PrimitivePatch<Face, FaceList, PointField>::"
            "calcMeshData()"
        )   << "meshPointsPtr_ or localFacesPtr_already allocated"
            << abort(FatalError);
    }

    // Create a map for marking points.  Estimated size is 4 times the
    // number of faces in the patch
    Map<label> markedPoints(4*this->size());

    // if the point is used, set the mark to 1
    forAll (*this, faceI)
    {
        const Face& curPoints = this->operator[](faceI);

        forAll (curPoints, pointI)
        {
            markedPoints.insert(curPoints[pointI], -1);
        }
    }

    // Create the storage and store the meshPoints.  Mesh points are
    // the ones marked by the usage loop above
    meshPointsPtr_ = new labelList(markedPoints.toc());
    labelList& pointPatch = *meshPointsPtr_;

    // Sort the list to preserve compatibility with the old ordering
    sort(pointPatch);

    // For every point in map give it its label in mesh points
    forAll (pointPatch, pointI)
    {
        markedPoints.find(pointPatch[pointI])() = pointI;
    }

    // Create local faces. Note that we start off from copy of original face
    // list (even though vertices are overwritten below). This is done so
    // additional data gets copied (e.g. region number of labelledTri)
    localFacesPtr_ = new List<Face>(*this);
    List<Face>& lf = *localFacesPtr_;

    forAll (*this, faceI)
    {
        const Face& curFace = this->operator[](faceI);
        lf[faceI].setSize(curFace.size());

        forAll (curFace, labelI)
        {
            lf[faceI][labelI] = markedPoints.find(curFace[labelI])();
        }
    }

    if (debug)
    {
        Info<< "PrimitivePatch<Face, FaceList, PointField>::"
               "calcMeshData() : "
               "finished calculating mesh data in PrimitivePatch"
            << endl;
    }
}


template<class Face, template<class> class FaceList, class PointField>
void PrimitivePatch<Face, FaceList, PointField>::calcMeshPointMap() const
{
    if (debug)
    {
        Info<< "PrimitivePatch<Face, FaceList, PointField>::"
               "calcMeshPointMap() : "
               "calculating mesh point map in PrimitivePatch"
            << endl;
    }

    // It is considered an error to attempt to recalculate meshPoints
    // if they have already been calculated.
    if (meshPointMapPtr_)
    {
        FatalErrorIn
        (
            "PrimitivePatch<Face, FaceList, PointField>::"
            "calcMeshPointMap()"
        )   << "meshPointMapPtr_ already allocated"
            << abort(FatalError);
    }

    const labelList& mp = meshPoints();

    meshPointMapPtr_ = new Map<label>(2*mp.size());
    Map<label>& mpMap = *meshPointMapPtr_;

    forAll (mp, i)
    {
        mpMap.insert(mp[i], i);
    }

    if (debug)
    {
        Info<< "PrimitivePatch<Face, FaceList, PointField>::"
               "calcMeshPointMap() : "
               "finished calculating mesh point map in PrimitivePatch"
            << endl;
    }
}


template<class Face, template<class> class FaceList, class PointField>
void PrimitivePatch<Face, FaceList, PointField>::calcLocalPoints() const
{
    if (debug)
    {
        Info<< "PrimitivePatch<Face, FaceList, PointField>::"
               "calcLocalPoints() : "
               "calculating localPoints in PrimitivePatch"
            << endl;
    }

    // It is considered an error to attempt to recalculate localPoints
    // if they have already been calculated.
    if (localPointsPtr_)
    {
        FatalErrorIn
        (
            "PrimitivePatch<Face, FaceList, PointField>::calcLocalPoints()"
        )   << "localPointsPtr_already allocated"
            << abort(FatalError);
    }

    const labelList& meshPts = meshPoints();

    localPointsPtr_ = new pointField(meshPts.size());

    pointField& locPts = *localPointsPtr_;

    forAll (meshPts, pointI)
    {
        locPts[pointI] = points_[meshPts[pointI]];
    }

    if (debug)
    {
        Info<< "PrimitivePatch<Face, FaceList, PointField>::"
            << "calcLocalPoints() : "
            << "finished calculating localPoints in PrimitivePatch"
            << endl;
    }
}


template<class Face, template<class> class FaceList, class PointField>
void PrimitivePatch<Face, FaceList, PointField>::calcPointNormals() const
{
    if (debug)
    {
        Info<< "PrimitivePatch<Face, FaceList, PointField>::"
               "calcPointNormals() : "
               "calculating pointNormals in PrimitivePatch"
            << endl;
    }

    // It is considered an error to attempt to recalculate pointNormals
    // if they have already been calculated.
    if (pointNormalsPtr_)
    {
        FatalErrorIn
        (
            "PrimitivePatch<Face, FaceList, PointField>::calcPointNormals()"
        )   << "pointNormalsPtr_already allocated"
            << abort(FatalError);
    }

    const vectorField& faceUnitNormals = faceNormals();

    const labelListList& pf = pointFaces();

    pointNormalsPtr_ = new vectorField(meshPoints().size(), vector(0, 0, 0));

    vectorField& n = *pointNormalsPtr_;

    forAll (pf, pointI)
    {
        vector& curNormal = n[pointI];

        const labelList& curFaces = pf[pointI];

        forAll (curFaces, faceI)
        {
            curNormal += faceUnitNormals[curFaces[faceI]];
        }

        curNormal /= mag(curNormal) + VSMALL;
    }

    if (debug)
    {
        Info<< "PrimitivePatch<Face, FaceList, PointField>::"
               "calcPointNormals() : "
               "finished calculating pointNormals in PrimitivePatch"
            << endl;
    }
}


template<class Face, template<class> class FaceList, class PointField>
void PrimitivePatch<Face, FaceList, PointField>::calcFaceNormals() const
{
    if (debug)
    {
        Info<< "PrimitivePatch<Face, FaceList, PointField>::"
               "calcFaceNormals() : "
               "calculating faceNormals in PrimitivePatch"
            << endl;
    }

    // It is considered an error to attempt to recalculate faceNormals
    // if they have already been calculated.
    if (faceNormalsPtr_)
    {
        FatalErrorIn
        (
            "PrimitivePatch<Face, FaceList, PointField>::calcFaceNormals()"
        )   << "faceNormalsPtr_already allocated"
            << abort(FatalError);
    }

    faceNormalsPtr_ = new vectorField(this->size());

    vectorField& n = *faceNormalsPtr_;

    forAll (n, faceI)
    {
        n[faceI] = this->operator[](faceI).normal(points_);
    }

    // Normalize the normals
    n /= mag(n) + VSMALL;

    if (debug)
    {
        Info<< "PrimitivePatch<Face, FaceList, PointField>::"
               "calcFaceNormals() : "
               "finished calculating faceNormals in PrimitivePatch"
            << endl;
    }
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
