/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Class
    directPolyTopoChange

Description
    Direct mesh changes based on polyTopoChange syntax.

    Instead of recording changes and executing them all in one go (as does
    polyTopoChange) this class actually holds the current points/faces/cells
    and does the change immediately.
    It can be asked to compress out all unused points/faces/cells and
    renumber everything to be consistent.

    Note:
    - adding a face using non-existing cells causes all intermediate cells
    to be added. So always first add cells/points and then faces.
    (or set strict checking)
    - strict checking:
        -any added/modified face can only use already existing vertices
        -any added face                 ,,                     cells
        -no item can be removed more than once.
    - removed cell: cell set to 0 faces.
    - removed face: face set to 0 vertices.
    - removed point: coordinate set to greatPoint (GREAT,GREAT,GREAT).
    Note that this might give problems if this value is used already.
    To see if point is equal to above value we don't use == (which might give
    problems with roundoff error) but instead compare the individual component
    with >.
    - coupled patches: the reorderCoupledFaces routine (borrowed from
    the couplePatches utility) reorders coupled patch faces and
    uses the cyclicPolyPatch,processorPolyPatch functionality.

SourceFiles
    directPolyTopoChange.C
    directPolyTopoChangeI.H
    directPolyTopoChangeTemplates.C

\*---------------------------------------------------------------------------*/

#ifndef directPolyTopoChange_H
#define directPolyTopoChange_H

#include "autoPtr.H"
#include "DynamicList.H"
#include "labelList.H"
#include "IOobject.H"
#include "typeInfo.H"
#include "pointField.H"
#include "PtrList.H"
#include "cellList.H"
#include "Map.H"
#include "labelHashSet.H"
#include "mapPolyMesh.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// Class forward declarations
class face;
class primitiveMesh;
class polyMesh;
class fvMesh;
class Time;
class fileName;
class polyBoundaryMesh;
class polyPatch;
class dictionary;
class topoAction;
class objectMap;

/*---------------------------------------------------------------------------*\
                           Class directPolyTopoChange Declaration
\*---------------------------------------------------------------------------*/

class directPolyTopoChange
{
        ////- Less function class used by upper-triangular face ordering
        ////  - internal faces according to owner or for same owner according to
        ////    neighbour
        ////  - boundary faces according to region
        //class faceLess
        //{
        //    const unallocLabelList& own_;
        //    const unallocLabelList& nei_;
        //    const unallocLabelList& region_;
        //
        //public:
        //
        //    faceLess
        //    (
        //        const unallocLabelList& own,
        //        const unallocLabelList& nei,
        //        const unallocLabelList& region
        //    )
        //    :
        //        own_(own),
        //        nei_(nei),
        //        region_(region)
        //    {}
        //
        //    bool operator()(const label a, const label b)
        //    {
        //        if (nei_[a] == -1)
        //        {
        //            if (nei_[b] == -1)
        //            {
        //                // Two boundary faces
        //                return region_[a] < region_[b];
        //            }
        //            else
        //            {
        //                // b is internal face.
        //                return false;
        //            }
        //        }
        //        else
        //        {
        //            if (nei_[b] == -1)
        //            {
        //                return true;
        //            }
        //            else
        //            {
        //                // two internal faces.
        //                if (own_[a] < own_[b])
        //                {
        //                    return true;
        //                }
        //                else if (own_[a] > own_[b])
        //                {
        //                    return false;
        //                }
        //                else
        //                {
        //                    return nei_[a] < nei_[b];
        //                }
        //            }
        //        }
        //    }
        //};



private:

    // Static data members

        //- Value of deleted point
        static const point greatPoint;


    // Private data

        //- Whether to allow referencing illegal points/cells/faces
        //  when adding/removing data.
        const bool strict_;


        // Patches

            //- Number of patches
            label nPatches_;


        // Points

            //- Current point set
            DynamicList<point> points_;

            //- Original point label (or masterpoint for added points)
            DynamicList<label> pointMap_;

            //- For all original and added points contains new point label.
            //  (used to map return value of addPoint to new mesh point)
            DynamicList<label> reversePointMap_;

            //- Zone of point
            Map<label> pointZone_;

            //- Retired points
            labelHashSet retiredPoints_;


        // Faces

            //- Current faceList
            DynamicList<face> faces_;

            //- Patch for every external face (-1 for internal faces)
            DynamicList<label> region_;

            //- Owner for all faces
            DynamicList<label> faceOwner_;

            //- Neighbour for internal faces (-1 for external faces)
            DynamicList<label> faceNeighbour_;

            //- Original face label. Or master face for added-from-faces; 
            //  -1 for faces added-from-edge or added-from-point)
            DynamicList<label> faceMap_;

            //- For all original and added faces contains new face label
            //  (used to map return value of addFace to new mesh face)
            DynamicList<label> reverseFaceMap_;

            //- Faces added from point (corresponding faceMap_ will
            //  be -1)
            Map<label> faceFromPoint_;

            //- Faces added from edge (corresponding faceMap_ will
            //  be -1)
            Map<label> faceFromEdge_;

            //- In mapping whether to reverse the flux.
            labelHashSet flipFaceFlux_;

            //- Zone of face
            Map<label> faceZone_;

            //- Orientation of face in zone
            Map<bool> faceZoneFlip_;

            //- Active faces
            label nActiveFaces_;


        // Cells

            //- Original cell label or master cell for added-from-cell;
            //  -1 for cells added from face or edge.
            DynamicList<label> cellMap_;

            //- For all original and added cells contains new cell label
            //  (used to map return value of addCell to new mesh cell)
            DynamicList<label> reverseCellMap_;

            //- Cells added from point
            Map<label> cellFromPoint_;

            //- Cells added from edge
            Map<label> cellFromEdge_;

            //- Cells added from face
            Map<label> cellFromFace_;

            //- Zone of cell
            DynamicList<label> cellZone_;


    // Private Member Functions

        //- Reorder contents of container according to map
        template<class T>
        static void reorder(const labelList& map, DynamicList<T>&);
        template<class T>
        static void reorder(const labelList& map, List<DynamicList<T> >&);
        template<class T>
        static void renumberKey(const labelList& map, Map<T>&);

        //- Renumber elements of list according to map
        static void renumber(const labelList&, DynamicList<label>&);
        static void renumber(const labelList&, labelHashSet&);

        //- Renumber & compact elements of list according to map
        static void renumberCompact(const labelList&, labelList&);

        //- Remove selected element from labelList
        //static void remove(const label elemToRemove, labelList&);

        //- Is point removed?
        inline bool pointRemoved(const label pointI) const;
        inline bool faceRemoved(const label faceI) const;
        inline bool cellRemoved(const label cellI) const;

        //- Check inputs to modFace or addFace
        void checkFace
        (
            const face&,
            const label faceI,
            const label own,
            const label nei,
            const label patchI,
            const label zoneI
        ) const;

        //- Construct cells (in packed storage)
        void makeCells
        (
            const label nActiveFaces,
            labelList& cellFaces,
            labelList& cellFaceOffsets
        ) const;

        //- Do upper-triangular ordering and patch ordering.
        labelList getFaceOrder
        (
            const label nActiveFaces,
            const labelList& cellFaces,
            const labelList& cellFaceOffsets
        ) const;

        //- Compact and reorder faces according to map
        void reorderCompactFaces
        (
            const label newSize,
            const labelList& oldToNew
        );

        //- Remove all unused/removed points/faces/cells and update
        //  face ordering.
        void compact();

        //- Given a sorted faceList calculates sizes and starts of patches.
        void calcPatchSizes
        (
            labelList& patchSizes,
            labelList& patchStarts
        ) const;

        //- Select either internal or external faces out of faceLabels
        static labelList selectFaces
        (
            const primitiveMesh&,
            const labelList& faceLabels,
            const bool internalFacesOnly
        );

        //- Calculate mapping for patchpoints only
        void calcPatchPointMap
        (
            const List<Map<label> >&,
            const labelList&,
            const polyBoundaryMesh&,
            labelListList&
        ) const;

        void calcFaceInflationMaps
        (
            const polyMesh&,
            List<objectMap>&,
            List<objectMap>&
        ) const;

        void calcCellInflationMaps
        (
            const polyMesh&,
            List<objectMap>&,
            List<objectMap>&,
            List<objectMap>& 
        ) const;

        void resetZones
        (
            polyMesh&,
            labelListList&,
            labelListList&,
            labelListList&
        ) const;

        void calcFaceZonePointMap
        (
            polyMesh&,
            const List<Map<label> >&,
            labelListList&
        ) const;


        // Coupling

            //- Rotate face by number of positions
            static face rotateFace(const face& f, const label nPos);

            //- Do all coupled patch face reordering
            void reorderCoupledFaces
            (
                const polyBoundaryMesh&,
                const labelList& patchStarts,
                const labelList& patchSizes,
                const pointField& points
            );

public:

    //- Runtime type information
    ClassName("directPolyTopoChange");



    // Constructors

        //- Construct from mesh. Adds all points/face/cells from mesh.
        directPolyTopoChange(const polyMesh& mesh, const bool strict = true);


    // Member Functions

        // Access

            //- Points. Shrunk after constructing mesh (or calling of compact())
            const DynamicList<point>& points() const
            {
                return points_;
            }

            const DynamicList<face>& faces() const
            {
                return faces_;
            }

            const DynamicList<label>& region()const
            {
                return region_;
            }

            const DynamicList<label>& faceOwner()const
            {
                return faceOwner_;
            }

            const DynamicList<label>& faceNeighbour()const
            {
                return faceNeighbour_;
            }


        // Edit

            //- Clear all storage
            void clear();

            //- Add all points/faces/cells of mesh. Additional offset for patch
            //  or zone ids.
            void addMesh
            (
                const polyMesh&,
                const labelList& patchMap,
                const labelList& pointZoneMap,
                const labelList& faceZoneMap,
                const labelList& cellZoneMap
            );


            //- For compatibility with polyTopoChange: set topological action.
            label setAction(const topoAction& action);

            //- Add point. Return new point label.
            //  Notes:
            //  - masterPointID can be < 0 (appended points)
            //  - inCell = false: add retired point (to end of point list)
            label addPoint
            (
                const point&,
                const label masterPointID,
                const label zoneID,
                const bool inCell
            );

            //- Modify coordinate.
            //  Notes:
            //  - inCell = false: add retired point (to end of point list)
            void modifyPoint
            (
                const label,
                const point&,
                const label newZoneID,
                const bool inCell
            );

            //- Remove point.
            void removePoint(const label);

            //- Add face to cells. Return new face label.
            //  own,nei<0, zoneID>=0 : add inactive face (to end of face list)
            label addFace
            (
                const face& f,
                const label own,
                const label nei,
                const label masterPointID,
                const label masterEdgeID,
                const label masterFaceID,
                const bool flipFaceFlux,
                const label patchID,
                const label zoneID,
                const bool zoneFlip
            );

            //- Modify vertices or cell of face.
            void modifyFace
            (
                const face& f,
                const label faceI,
                const label own,
                const label nei,
                const bool flipFaceFlux,
                const label patchID,
                const label zoneID,
                const bool zoneFlip
            );

            //- Remove face.
            void removeFace(const label);

            //- Add cell. Return new cell label.
            label addCell
            (
                const label masterPointID,
                const label masterEdgeID,
                const label masterFaceID,
                const label masterCellID,
                const label zoneID
            );

            //- Modify zone of cell
            void modifyCell(const label, const label zoneID);

            //- Remove cell.
            void removeCell(const label);


        // Other

            //- Inplace changes mesh without change of patches.
            //  Adapts patch start/end and does reorderCoupledFaces.
            //  Clears all data. Returns map.
            autoPtr<mapPolyMesh> changeMesh(polyMesh& mesh);
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#include "directPolyTopoChangeI.H"

#ifdef NoRepository
#   include "directPolyTopoChangeTemplates.C"
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
