/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description

\*---------------------------------------------------------------------------*/

#include "faMesh.H"
#include "lduAddressingFaMesh.H"
#include "dimensionSet.H"
#include "areaFields.H"
#include "edgeFields.H"
#include "primitiveFacePatch.H"
#include "fac.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void faMesh::calcLduAddressing() const
{
    if (debug)
    {
        Info<< "void faMesh::calcLduAddressing() const : "
            << "Calculating addressing" << endl;
    }

    if (lduPtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcLduAddressing() const"
        )   << "lduPtr_ already allocated"
            << abort(FatalError);
    }

    lduPtr_ = new lduAddressingFaMesh(*this);
}


void faMesh::calcPatchStarts() const
{
    if (debug)
    {
        Info<< "void faMesh::calcPatchStarts() const : "
            << "Calculating patch starts" << endl;
    }

    if (patchStartsPtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcPatchStarts() const"
        )   << "patchStartsPtr_ already allocated"
            << abort(FatalError);
    }

    patchStartsPtr_ = new labelList(boundary().size(), -1);
    labelList& patchStarts = *patchStartsPtr_;

    patchStarts[0] = nInternalEdges();

    for(label i=1; i<boundary().size(); i++)
    {
        patchStarts[i] =
            patchStarts[i-1]
          + boundary()[i-1].size();
    }
}


void faMesh::calcLe() const
{
    if (debug)
    {
        Info<< "void faMesh::calcLe() const : "
            << "Calculating local edges" << endl;
    }

    if (LePtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcLe() const"
        )   << "LePtr_ already allocated"
            << abort(FatalError);
    }

    LePtr_ =
        new edgeVectorField
        (
            IOobject
            (
                "Le",
                mesh_.meshDir(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE,
                false
            ),
            *this,
            dimLength
        );

    edgeVectorField& Le = *LePtr_;


    const pointField& pPoints = points();
    const edgeList& pEdges = edges();

    const edgeVectorField& eCentres = edgeCentres();
    const areaVectorField& fCentres = centres();

    const edgeVectorField& edgeNormals = edgeAreaNormals();

    vectorField& leInternal = Le.internalField();
    const vectorField& edgeNormalsInternal = edgeNormals.internalField();
    const vectorField& fCentresInternal = fCentres.internalField();
    const vectorField& eCentresInternal = eCentres.internalField();
    const scalarField& magLeInternal = magLe().internalField();

    forAll (leInternal, edgeI)
    {
        leInternal[edgeI] =
            pEdges[edgeI].vec(pPoints) ^ edgeNormalsInternal[edgeI];

        leInternal[edgeI] *=
          - sign
            (
                leInternal[edgeI] &
                (
                    fCentresInternal[owner()[edgeI]]
                  - eCentresInternal[edgeI]
                )
            );

        leInternal[edgeI] *= 
            magLeInternal[edgeI]/mag(leInternal[edgeI]);
    }


    forAll (boundary(), patchI)
    {
        const labelList::subList bndEdgeFaces =
            boundary()[patchI].edgeFaces();

        const edgeList::subList bndEdges =
            boundary()[patchI].patchSlice(pEdges);

        const vectorField& bndEdgeNormals =
            edgeNormals.boundaryField()[patchI];

        vectorField& patchLe = Le.boundaryField()[patchI];
        const vectorField& patchECentres = eCentres.boundaryField()[patchI];

        forAll (patchLe, edgeI)
        {
            patchLe[edgeI] =
                bndEdges[edgeI].vec(pPoints) ^ bndEdgeNormals[edgeI];

            patchLe[edgeI] *=
              - sign
                (
                    patchLe[edgeI]&
                    (
                        fCentresInternal[bndEdgeFaces[edgeI]]
                      - patchECentres[edgeI]
                    )
                );

            patchLe[edgeI] *= 
                magLe().boundaryField()[patchI][edgeI]
                /mag(patchLe[edgeI]);
        }
    }
}


void faMesh::calcMagLe() const
{
    if (debug)
    {
        Info<< "void faMesh::calcMagLe() const : "
            << "Calculating local edge magnitudes" << endl;
    }

    if (magLePtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcMagLe() const"
        )   << "magLePtr_ already allocated"
            << abort(FatalError);
    }

    magLePtr_ =
        new edgeScalarField
        (
            IOobject
            (
                "magLe",
                mesh_.meshDir(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE,
                false
            ),
            *this,
            dimLength
        );

    edgeScalarField& magLe = *magLePtr_;


    const pointField& localPoints = points();

    const edgeList::subList internalEdges = 
        edgeList::subList(edges(), nInternalEdges());


    forAll (internalEdges, edgeI)
    {
        magLe.internalField()[edgeI] =
            internalEdges[edgeI].mag(localPoints);
    }


    forAll (boundary(), patchI)
    {
        const edgeList::subList patchEdges = 
            boundary()[patchI].patchSlice(edges());

        forAll (patchEdges, edgeI)
        {
            magLe.boundaryField()[patchI][edgeI] =
                patchEdges[edgeI].mag(localPoints);
        }
    }
}

void faMesh::calcCentres() const
{
    if (debug)
    {
        Info<< "void faMesh::calcCentres() const : "
            << "Calculating face centres" << endl;
    }

    if (centresPtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcCentres() const"
        )   << "centresPtr_ already allocated"
            << abort(FatalError);
    }

    centresPtr_ =
        new areaVectorField
        (
            IOobject
            (
                "centres",
                mesh_.meshDir(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE,
                false
            ),
            *this,
            dimLength
        );

    areaVectorField& centres = *centresPtr_;

    const pointField& localPoints = points();
    const faceList& localFaces = faces();


    forAll (localFaces, faceI)
    {
        centres.internalField()[faceI] =
            localFaces[faceI].centre(localPoints);
    }


    forAll (boundary(), patchI)
    {
        const edgeList::subList patchEdges = 
            boundary()[patchI].patchSlice(edges());

        forAll (patchEdges, edgeI)
        {
            centres.boundaryField()[patchI][edgeI] =
                patchEdges[edgeI].centre(localPoints);
        }
    }
}


void faMesh::calcEdgeCentres() const
{
    if (debug)
    {
        Info<< "void faMesh::calcEdgeCentres() const : "
            << "Calculating edge centres" << endl;
    }

    if (edgeCentresPtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcEdgeCentres() const"
        )   << "edgeCentresPtr_ already allocated"
            << abort(FatalError);
    }

    edgeCentresPtr_ =
        new edgeVectorField
        (
            IOobject
            (
                "edgeCentres",
                mesh_.meshDir(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE,
                false
            ),
            *this,
            dimLength
        );

    edgeVectorField& edgeCentres = *edgeCentresPtr_;

    const pointField& localPoints = points();

    const edgeList::subList internalEdges = 
        edgeList::subList(edges(), nInternalEdges());


    forAll (internalEdges, edgeI)
    {
        edgeCentres.internalField()[edgeI] =
            internalEdges[edgeI].centre(localPoints);
    }


    forAll (boundary(), patchI)
    {
        const edgeList::subList patchEdges = 
            boundary()[patchI].patchSlice(edges());

        forAll (patchEdges, edgeI)
        {
            edgeCentres.boundaryField()[patchI][edgeI] =
                patchEdges[edgeI].centre(localPoints);
        }
    }
}


void faMesh::calcS() const
{
    if (debug)
    {
        Info<< "void faMesh::calcS() const : "
            << "Calculating areas" << endl;
    }

    if (SPtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcS() const"
        )   << "SPtr_ already allocated"
            << abort(FatalError);
    }
    
    SPtr_ = new scalarField(nFaces(), 0.0);
    scalarField& S = *SPtr_;

    const pointField& localPoints = points();
    const faceList& localFaces = faces();

    forAll (S, faceI)
    {
        S[faceI] = localFaces[faceI].mag(localPoints);
    }
}


void faMesh::calcFaceAreaNormals() const
{
    if (debug)
    {
        Info<< "void faMesh::calcFaceAreaNormals() const : "
            << "Calculating face area normals" << endl;
    }

    if (faceAreaNormalsPtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcFaceAreaNormals() const"
        )   << "faceAreaNormalsPtr_ already allocated"
            << abort(FatalError);
    }

    faceAreaNormalsPtr_ =
        new areaVectorField
        (
            IOobject
            (
                "faceAreaNormals",
                mesh_.meshDir(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE,
                false
            ),
            *this,
            dimless
        );

    areaVectorField& faceAreaNormals = *faceAreaNormalsPtr_;

    const pointField& localPoints = points();
    const faceList& localFaces = faces();

    forAll (localFaces, faceI)
    {
        faceAreaNormals.internalField()[faceI] =
            localFaces[faceI].normal(localPoints)/
            localFaces[faceI].mag(localPoints);
    }

    forAll (boundary(), patchI)
    {
        faceAreaNormals.boundaryField()[patchI] =
            edgeAreaNormals().boundaryField()[patchI];
    }
}


void faMesh::calcEdgeAreaNormals() const
{
    if (debug)
    {
        Info<< "void faMesh::calcEdgeAreaNormals() const : "
            << "Calculating edge area normals" << endl;
    }

    if (edgeAreaNormalsPtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcEdgeAreaNormals() const"
        )   << "edgeAreaNormalsPtr_ already allocated"
            << abort(FatalError);
    }

    edgeAreaNormalsPtr_ =
        new edgeVectorField
        (
            IOobject
            (
                "edgeAreaNormals",
                mesh_.meshDir(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE,
                false
            ),
            *this,
            dimless
        );

    edgeVectorField& edgeAreaNormals = *edgeAreaNormalsPtr_;


    // Point area normals
    const vectorField& pointNormals = pointAreaNormals();


//     // Primitive patch edge normals
//     const labelListList& patchPointEdges = patch().pointEdges();

//     vectorField patchEdgeNormals(nEdges(), vector::zero);

//     forAll(pointNormals, pointI)
//     {
//         const labelList& curPointEdges = patchPointEdges[pointI];

//         forAll(curPointEdges, edgeI)
//         {
//             label curEdge = curPointEdges[edgeI];

//             patchEdgeNormals[curEdge] += 0.5*pointNormals[pointI];
//         }
//     }

//     patchEdgeNormals /= mag(patchEdgeNormals);


//     // Edge area normals
//     label nIntEdges = patch().nInternalEdges();

//     for (label edgeI = 0; edgeI < nIntEdges; edgeI++)
//     {
//         edgeAreaNormals.internalField()[edgeI] =
//             patchEdgeNormals[edgeI];
//     }
    
//     forAll (boundary(), patchI)
//     {
//         const labelList& edgeLabels = boundary()[patchI];

//         forAll (edgeAreaNormals.boundaryField()[patchI], edgeI)
//         {
//             edgeAreaNormals.boundaryField()[patchI][edgeI] =
//                 patchEdgeNormals[edgeLabels[edgeI]];
//         }
//     }


    forAll (edgeAreaNormals.internalField(), edgeI)
    {
        vector e = edges()[edgeI].vec(points());
        
        e /= mag(e);

        

//         scalar wStart = 
//             1.0 - sqr(mag(e^pointNormals[edges()[edgeI].end()]));

//         scalar wEnd = 
//             1.0 - sqr(mag(e^pointNormals[edges()[edgeI].start()]));

//         wStart = 1.0;
//         wEnd = 1.0;

//         edgeAreaNormals.internalField()[edgeI] =
//             wStart*pointNormals[edges()[edgeI].start()]
//           + wEnd*pointNormals[edges()[edgeI].end()];

//         vector eC = 0.5*(points()[edges()[edgeI].start()] + points()[edges()[edgeI].end()]);

//         vector eCp = 0.5*
//             (
//                 points()[edges()[edgeI].start()] + pointNormals[edges()[edgeI].start()]
//                 points()[edges()[edgeI].end()] + 
//             );

        edgeAreaNormals.internalField()[edgeI] =
            pointNormals[edges()[edgeI].start()] + pointNormals[edges()[edgeI].end()];

        edgeAreaNormals.internalField()[edgeI] -= e*(e&edgeAreaNormals.internalField()[edgeI]);
    }

    
    edgeAreaNormals.internalField() /=
        mag(edgeAreaNormals.internalField());
    
    

    forAll (boundary(), patchI)
    {
        const edgeList::subList patchEdges =
            boundary()[patchI].patchSlice(edges());

        forAll (patchEdges, edgeI)
        {
            edgeAreaNormals.boundaryField()[patchI][edgeI] =
                pointNormals[patchEdges[edgeI].start()]
              + pointNormals[patchEdges[edgeI].end()];

            vector e = patchEdges[edgeI].vec(points());
            e /= mag(e);

            edgeAreaNormals.boundaryField()[patchI][edgeI] -= 
                e*(e&edgeAreaNormals.boundaryField()[patchI][edgeI]);
        }
        
        edgeAreaNormals.boundaryField()[patchI] /=
            mag(edgeAreaNormals.boundaryField()[patchI]);
    }
}


void faMesh::calcFaceCurvatures() const
{
    if (debug)
    {
        Info<< "void faMesh::calcFaceCurvatures() const : "
            << "Calculating face curvatures" << endl;
    }

    if (faceCurvaturesPtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcFaceCurvatures() const"
        )   << "faceCurvaturesPtr_ already allocated"
            << abort(FatalError);
    }

    faceCurvaturesPtr_ =
        new areaScalarField
        (
            IOobject
            (
                "faceCurvatures",
                mesh_.meshDir(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE,
                false
            ),
            *this,
            dimless/dimLength
        );

    areaScalarField& faceCurvatures = *faceCurvaturesPtr_;


//     faceCurvatures =
//         fac::edgeIntegrate(Le()*edgeLengthCorrection())
//         &faceAreaNormals();

    areaVectorField kN =
        fac::edgeIntegrate(Le()*edgeLengthCorrection());

    faceCurvatures = sign(kN&faceAreaNormals())*mag(kN);
}


void faMesh::calcEdgeTransformTensors() const
{
    if (debug)
    {
        Info<< "void faMesh::calcEdgeTransformTensors() const : "
            << "Calculating edge transformation tensors" << endl;
    }

    if (edgeTransformTensorsPtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcEdgeTransformTensors() const"
        )   << "edgeTransformTensorsPtr_ already allocated"
            << abort(FatalError);
    }

    edgeTransformTensorsPtr_ = new FieldField<Field, tensor>(nEdges());
    FieldField<Field, tensor>& edgeTransformTensors =
        *edgeTransformTensorsPtr_;

    const edgeVectorField& Ne = edgeAreaNormals();
    const areaVectorField& Nf = faceAreaNormals();
    const areaVectorField& Cf = centres();
    const edgeVectorField& Ce = edgeCentres();


    // Internal edges transformation tensors
    for(label edgeI=0; edgeI<nInternalEdges(); edgeI++)
    {
        edgeTransformTensors.hook(new Field<tensor>(3, I));

        vector E = Ce.internalField()[edgeI];

        if (skew())
        {
            E -= skewCorrectionVectors().internalField()[edgeI];
        }

        // Edge transformation tensor
        vector il = E - Cf.internalField()[owner()[edgeI]];
        
        il -= Ne.internalField()[edgeI]
            *(Ne.internalField()[edgeI]&il);

        il /= mag(il);

        vector kl = Ne.internalField()[edgeI];
        vector jl = kl^il;

        edgeTransformTensors[edgeI][0] =
            tensor
            (
                il.x(), il.y(), il.z(),
                jl.x(), jl.y(), jl.z(),
                kl.x(), kl.y(), kl.z()
            );

        // Owner transformation tensor
        il = E - Cf.internalField()[owner()[edgeI]];

        il -= Nf.internalField()[owner()[edgeI]]
            *(Nf.internalField()[owner()[edgeI]]&il);

        il /= mag(il);

        kl = Nf.internalField()[owner()[edgeI]];
        jl = kl^il;

        edgeTransformTensors[edgeI][1] =
            tensor
            (
                il.x(), il.y(), il.z(),
                jl.x(), jl.y(), jl.z(),
                kl.x(), kl.y(), kl.z()
            );

        // Neighbour transformation tensor
        il = Cf.internalField()[neighbour()[edgeI]] - E;

        il -= Nf.internalField()[neighbour()[edgeI]]
            *(Nf.internalField()[neighbour()[edgeI]]&il);

        il /= mag(il);

        kl = Nf.internalField()[neighbour()[edgeI]];
        jl = kl^il;

        edgeTransformTensors[edgeI][2] =
            tensor
            (
                il.x(), il.y(), il.z(),
                jl.x(), jl.y(), jl.z(),
                kl.x(), kl.y(), kl.z()
            );
    }


    // Boundary edges transformation tensors
    forAll (boundary(), patchI)
    {
        const labelList::subList edgeFaces = boundary()[patchI].edgeFaces();

        forAll(edgeFaces, edgeI)
        {
            edgeTransformTensors.hook(new Field<tensor>(2, I));

            vector E = Ce.boundaryField()[patchI][edgeI];

            if (skew())
            {
                E -= skewCorrectionVectors().boundaryField()[patchI][edgeI];
            }

            // Edge transformation tensor
            vector il = E - Cf.internalField()[edgeFaces[edgeI]];
        
            il -= Ne.boundaryField()[patchI][edgeI]
                *(Ne.boundaryField()[patchI][edgeI]&il);

            il /= mag(il);

            vector kl = Ne.boundaryField()[patchI][edgeI];
            vector jl = kl^il;

            edgeTransformTensors[boundary()[patchI].start() + edgeI][0] = 
                tensor
                (
                    il.x(), il.y(), il.z(),
                    jl.x(), jl.y(), jl.z(),
                    kl.x(), kl.y(), kl.z()
                );

            // Owner transformation tensor
            il = E - Cf.internalField()[edgeFaces[edgeI]];

            il -= Nf.internalField()[edgeFaces[edgeI]]
                *(Nf.internalField()[edgeFaces[edgeI]]&il);

            il /= mag(il);

            kl = Nf.internalField()[edgeFaces[edgeI]];
            jl = kl^il;

            edgeTransformTensors[boundary()[patchI].start() + edgeI][1] = 
                tensor
                (
                    il.x(), il.y(), il.z(),
                    jl.x(), jl.y(), jl.z(),
                    kl.x(), kl.y(), kl.z()
                );
        }
    }
}


labelList faMesh::internalPoints() const
{
    if (debug)
    {
        Info<< "labelList faMesh::internalPoints() const : "
            << "Calculating internal points" << endl;
    }

    const edgeList& edges = patch().edges();
    label nIntEdges = patch().nInternalEdges();

    List<bool> internal (nPoints(), true);

    for (label curEdge = nIntEdges; curEdge < edges.size(); curEdge++)
    {
        internal[edges[curEdge].start()] = false;

        internal[edges[curEdge].end()] = false;
    }

    SLList<label> internalPoints;

    forAll(internal, pointI)
    {
        if(internal[pointI])
        {
            internalPoints.append(pointI);
        }
    }

    labelList result(internalPoints);

    return result;
}


labelList faMesh::boundaryPoints() const
{
    if (debug)
    {
        Info<< "labelList faMesh::boundaryPoints() const : "
            << "Calculating boundary points" << endl;
    }

    const edgeList& edges = patch().edges();
    label nIntEdges = patch().nInternalEdges();

    List<bool> internal (nPoints(), true);

    for (label curEdge = nIntEdges; curEdge < edges.size(); curEdge++)
    {
        internal[edges[curEdge].start()] = false;

        internal[edges[curEdge].end()] = false;
    }

    SLList<label> boundaryPoints;

    forAll(internal, pointI)
    {
        if(!internal[pointI])
        {
            boundaryPoints.append(pointI);
        }
    }

    labelList result(boundaryPoints);

    return result;
}


void faMesh::calcPointAreaNormals() const
{
    if (pointAreaNormalsPtr_)
    {
        FatalErrorIn
        (
            "void faMesh::calcPointAreaNormals() const"
        )   << "pointAreaNormalsPtr_ already allocated"
            << abort(FatalError);
    }


    pointAreaNormalsPtr_ =
        new vectorField
        (
            nPoints(),
            vector::zero
        );

    vectorField& result = *pointAreaNormalsPtr_;


    labelList intPoints = internalPoints();
    labelList bndPoints = boundaryPoints();

    const pointField& points = patch().localPoints();
    const faceList& faces = patch().localFaces();
    const labelListList& pointFaces = patch().pointFaces();

    forAll(intPoints, pointI)
    {
        label curPoint = intPoints[pointI];

        faceList curFaceList(pointFaces[curPoint].size());

        forAll(curFaceList, faceI)
        {
            curFaceList[faceI] = faces[pointFaces[curPoint][faceI]];
        }

        primitiveFacePatch curPatch (curFaceList, points);

        labelList curPointPoints = curPatch.edgeLoops()[0];

        for(int i=0; i<curPointPoints.size(); i++)
        {
            vector d1 =
                points[curPatch.meshPoints()[curPointPoints[i]]]
              - points[curPoint];

            label p = i+1;

            if(i == (curPointPoints.size()-1))
            {
                p = 0;
            }

            vector d2 =
                points[curPatch.meshPoints()[curPointPoints[p]]]
              - points[curPoint];

            vector n = (d1^d2)/mag(d1^d2);

            scalar sinAlpha = mag(d1^d2)/(mag(d1)*mag(d2));

            scalar w = sinAlpha/(mag(d1)*mag(d2));

            result[curPoint] += w*n;
        }
    }

    forAll(bndPoints, pointI)
    {
        label curPoint = bndPoints[pointI];

        faceList curFaceList(pointFaces[curPoint].size());

        forAll(curFaceList, faceI)
        {
            curFaceList[faceI] = faces[pointFaces[curPoint][faceI]];
        }

        primitiveFacePatch curPatch (curFaceList, points);

        labelList agglomFacePoints = curPatch.edgeLoops()[0];

        SLList<label> slList;

        label curPointLabel = -1;

        for(label i=0; i<agglomFacePoints.size(); i++)
        {
            if(curPatch.meshPoints()[agglomFacePoints[i]] == curPoint)
            {
                curPointLabel = i;
            }
            else if( curPointLabel != -1 )
            {
                slList.append(curPatch.meshPoints()[agglomFacePoints[i]]);
            }
        }

        for(label i=0; i<curPointLabel; i++)
        {
            slList.append(curPatch.meshPoints()[agglomFacePoints[i]]);
        }

        labelList curPointPoints(slList);

        for(label i=0; i<(curPointPoints.size()-1); i++)
        {
            vector d1 = points[curPointPoints[i]] - points[curPoint];

            vector d2 = points[curPointPoints[i+1]] - points[curPoint];

            vector n = (d1^d2)/mag(d1^d2);

            scalar sinAlpha = mag(d1^d2)/(mag(d1)*mag(d2));

            scalar w = sinAlpha/(mag(d1)*mag(d2));

            result[curPoint] += w*n;
        }
    }


    // Boundary points correction
    forAll (boundary(), patchI)
    {
        if(boundary()[patchI].ngbPolyPatchIndex() != -1)
        {
            labelList patchPoints = boundary()[patchI].pointLabels();
            vectorField N = boundary()[patchI].ngbPolyPatchPointNormals();

            forAll(patchPoints, pointI)
            {
                result[patchPoints[pointI]] 
                    -= N[pointI]*(N[pointI]&result[patchPoints[pointI]]);
            }
        }
    }
   

    result /= mag(result);
}


tmp<edgeScalarField> faMesh::edgeLengthCorrection() const
{
    if (debug)
    {
        Info<< "tmp<edgeScalarField> faMesh::edgeLengthCorrection() const : "
            << "Calculating edge length correction" << endl;
    }

    tmp<edgeScalarField> tcorrection
    (
        new edgeScalarField
        (
            IOobject
            (
                "edgeLengthCorrection",
                mesh_.meshDir(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE,
                false
            ),
            *this,
            dimless
        )
    );
    edgeScalarField& correction = tcorrection();


    const vectorField& pointNormals = pointAreaNormals();


    forAll (correction.internalField(), edgeI)
    {
        scalar sinAlpha = mag
        (
            pointNormals[edges()[edgeI].start()]^
            pointNormals[edges()[edgeI].end()]
        );

        scalar alpha = asin(sinAlpha);

        correction.internalField()[edgeI] = cos(alpha/2.0);
    }


    forAll (boundary(), patchI)
    {
        const edgeList::subList patchEdges =
            boundary()[patchI].patchSlice(edges());

        forAll (patchEdges, edgeI)
        {
            scalar sinAlpha = mag
                (
                    pointNormals[patchEdges[edgeI].start()]^
                    pointNormals[patchEdges[edgeI].end()]
                );

            scalar alpha = asin(sinAlpha);

            correction.boundaryField()[patchI][edgeI] = cos(alpha/2.0);
        }
    }


	return tcorrection;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
