/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Class
    linearUpwindV

Description
    linearUpwindV interpolation scheme class derived from upwind and returns
    upwind weighting factors but also applies an explicit correction.

SourceFiles
    linearUpwindV.C

\*---------------------------------------------------------------------------*/

#ifndef linearUpwindV_H
#define linearUpwindV_H

#include "upwind.H"
#include "gaussGrad.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

/*---------------------------------------------------------------------------*\
                           Class linearUpwindV Declaration
\*---------------------------------------------------------------------------*/

template<class Type>
class linearUpwindV
:
    public upwind<Type>
{
    // Private Data

        tmp<fv::gradScheme<Type> > gradScheme_;


    // Private Member Functions

        //- Disallow default bitwise copy construct
        linearUpwindV(const linearUpwindV&);

        //- Disallow default bitwise assignment
        void operator=(const linearUpwindV&);


public:

    //- Runtime type information
    TypeName("linearUpwindV");


    // Constructors

        //- Construct from faceFlux
        linearUpwindV
        (
            const fvMesh& mesh,
            const surfaceScalarField& faceFlux
        )
        :
            upwind<Type>(mesh, faceFlux),
            gradScheme_
            (
                new fv::gaussGrad<Type>(mesh)
            )
        {}

        //- Construct from Istream. 
        //  The name of the flux field is read from the Istream and looked-up
        //  from the mesh objectRegistry
        linearUpwindV
        (
            const fvMesh& mesh,
            Istream& schemeData
        )
        :
            upwind<Type>(mesh, schemeData),
            gradScheme_
            (
                fv::gradScheme<Type>::New
                (
                    mesh,
                    schemeData
                )
            )
        {}

        //- Construct from faceFlux and Istream
        linearUpwindV
        (
            const fvMesh& mesh,
            const surfaceScalarField& faceFlux,
            Istream& schemeData
        )
        :
            upwind<Type>(mesh, faceFlux, schemeData),
            gradScheme_
            (
                fv::gradScheme<Type>::New
                (
                    mesh,
                    schemeData
                )
            )
        {}


    // Member Functions

        //- Return true if this scheme uses an explicit correction
        virtual bool corrected() const
        {
            return true;
        }

        //- Return the explicit correction to the face-interpolate
        virtual tmp<GeometricField<Type, fvPatchField, surfaceMesh> >
        correction
        (
            const GeometricField<Type, fvPatchField, volMesh>& vf
        ) const
        {
            const fvMesh& mesh = this->mesh();

            tmp<GeometricField<Type, fvPatchField, surfaceMesh> > tsfCorr
            (
                new GeometricField<Type, fvPatchField, surfaceMesh>
                (
                    IOobject
                    (
                        vf.name(),
                        mesh.time().timeName(),
                        mesh
                    ),
                    mesh,
                    dimensioned<Type>
                    (
                        vf.name(),
                        vf.dimensions(),
                        pTraits<Type>::zero
                    )
                )
            );

            GeometricField<Type, fvPatchField, surfaceMesh>& sfCorr = tsfCorr();

            const surfaceScalarField& faceFlux = this->faceFlux_;
            const  surfaceScalarField& w = mesh.weights();

            const labelList& own = mesh.owner();
            const labelList& nei = mesh.neighbour();

            const vectorField& C = mesh.C();
            const vectorField& Cf = mesh.Cf();

            GeometricField
            <typename outerProduct<vector, Type>::type, fvPatchField, volMesh>
                gradVf = gradScheme_().grad(vf);

            forAll(faceFlux, facei)
            {
                vector maxCorr;

                if (faceFlux[facei] > 0.0)
                {
                    maxCorr =
                        (1.0 - w[facei])*(vf[nei[facei]] - vf[own[facei]]);

                    sfCorr[facei] = 
                        (Cf[facei] - C[own[facei]]) & gradVf[own[facei]];
                }
                else
                {
                    maxCorr =
                        w[facei]*(vf[own[facei]] - vf[nei[facei]]);

                    sfCorr[facei] = 
                        (Cf[facei] - C[nei[facei]]) & gradVf[nei[facei]];
                }

                scalar sfCorrs = magSqr(sfCorr[facei]);
                scalar maxCorrs = sfCorr[facei] & maxCorr;

                if (sfCorrs > 0)
                {
                    if (maxCorrs < 0)
                    {
                        sfCorr[facei] = vector::zero;
                    }
                    else if (sfCorrs > maxCorrs)
                    {
                        sfCorr[facei] *= maxCorrs/(sfCorrs + VSMALL);
                    }
                }
                else if (sfCorrs < 0)
                {
                    if (maxCorrs > 0)
                    {
                        sfCorr[facei] = vector::zero;
                    }
                    else if (sfCorrs < maxCorrs)
                    {
                        sfCorr[facei] *= maxCorrs/(sfCorrs - VSMALL);
                    }
                }
            }

            return tsfCorr;
        }
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
