/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

Class
    polyMeshAdder

Description
    Adds two meshes without using any polyMesh morphing. Gets faces to couple
    as faceCoupleInfo which is list of faces on both meshes. Holds map from
    last mesh addition.

SourceFiles
    polyMeshAdder.C

\*---------------------------------------------------------------------------*/

#ifndef polyMeshAdder_H
#define polyMeshAdder_H

#include "autoPtr.H"
#include "polyMesh.H"
#include "mapAddedPolyMesh.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// Class forward declarations
class IOobject;
class faceCoupleInfo;

/*---------------------------------------------------------------------------*\
                           Class polyMeshAdder Declaration
\*---------------------------------------------------------------------------*/

class polyMeshAdder
{

private:
    // Private data

        //- Current set of combined patches
        DynamicList<word> allPatchNames_;

        //- Current set of combined patch types.
        DynamicList<word> allPatchTypes_;


        //- After adding: map from new to original meshes.
        mapAddedPolyMesh map_;


    // Private Member Functions

        //- Append all elements of a list to a DynamicList
        template<class T>
        static void append(const List<T>&, DynamicList<T>&);

        //- Append all mapped elements of a list to a DynamicList
        static void append
        (
            const labelList& map,
            const labelList& lst,
            DynamicList<label>&
        );

        //- Return the vertices inbetween masterEdge endpoints in order.
        static labelList stringEdges
        (
            const pointField& cutPoints,
            const edge& meshEdge,
            const pointField& meshPoints,
            const labelList& cutToAllPoints,
            const edgeList& cutEdges,
            const labelList& cutEdgeLabels
        );

        //- Extract processor neighbour label from patchname
        static label neighbourProcNo(const word&);

        //- Index of patch in allPatches. Add if nonexisting.
        label patchIndex(const polyPatch&);

        //- Index of zone in all zones
        static label zoneIndex(const word&, DynamicList<word>&);

        //- Clone patch with new polyBoundaryMesh and size, start. Returns new
        //  patch index or -1 (if size was 0)
        static label clonePatch
        (
            const polyPatch& pp,
            const label start,
            const label size,
            const polyBoundaryMesh&,
            DynamicList<polyPatch*>&
        );

        //- Determine order for internalFaces to be upper-triangular.
        //  Does not change order of external faces.
        static labelList getFaceOrder
        (
            const cellList& cells,
            const label nInternalFaces,
            const labelList& owner,
            const labelList& neighbour
        );

        //- Inverts face-cell addressing to create cell-face addressing.
        static void makeCells
        (
            const label nCells,
            const labelList& owner,
            const labelList& neighbour,
            cellList& cells
        );

        //- Merge point zones
        static void mergePointZones
        (
            const pointZoneMesh& pz0,
            const pointZoneMesh& pz1,
            const labelList& from0ToAllPoints,
            const labelList& from1ToAllPoints,

            DynamicList<word>& zoneNames,
            labelList& from1ToAll,
            List<DynamicList<label> >& pzPoints
        );

        //- Merge face zones
        static void mergeFaceZones
        (
            const faceZoneMesh& fz0,
            const faceZoneMesh& fz1,
            const labelList& from0ToAllFaces,
            const labelList& from1ToAllFaces,

            DynamicList<word>& zoneNames,
            labelList& from1ToAll,
            List<DynamicList<label> >& fzFaces,
            List<DynamicList<bool> >& fzFlips
        );

        //- Merge cell zones
        static void mergeCellZones
        (
            const cellZoneMesh& cz0,
            const cellZoneMesh& cz1,
            const labelList& from1ToAllCells,

            DynamicList<word>& zoneNames,
            labelList& from1ToAll,
            List<DynamicList<label> >& czCells
        );

        void mergePatchNames
        (
            const polyBoundaryMesh& patches0,
            const polyBoundaryMesh& patches1,

            labelList& from1ToAllPatches,
            labelList& fromAllTo1Patches
        );

        //- Merges primitives of two meshes.
        void mergePrimitives
        (
            const polyMesh& mesh0,
            const polyMesh& mesh1,
            const faceCoupleInfo& coupleInfo,

            const labelList& fromAllTo1Patches,
            const labelList& from1ToAllPatches,

            pointField& allPoints,
            labelList& from0ToAllPoints,
            labelList& from1ToAllPoints,

            faceList& allFaces,
            labelList& from0ToAllFaces,
            labelList& from1ToAllFaces,
            labelList& nFaces,

            cellList& allCells,
            labelList& from1ToAllCells
        ) const;

        //- Merge point/face/cell zone information
        static void mergeZones
        (
            const polyMesh& mesh0,
            const polyMesh& mesh1,
            const labelList& from0ToAllPoints,
            const labelList& from0ToAllFaces,
            const labelList& from1ToAllPoints,
            const labelList& from1ToAllFaces,
            const labelList& from1ToAllCells,

            DynamicList<word>& pointZoneNames,
            List<DynamicList<label> >& pzPoints,

            DynamicList<word>& faceZoneNames,
            List<DynamicList<label> >& fzFaces,
            List<DynamicList<bool> >& fzFlips,

            DynamicList<word>& cellZoneNames,
            List<DynamicList<label> >& czCells
        );

        List<polyPatch*> combinePatches
        (
            const polyMesh& mesh0,
            const polyMesh& mesh1,
            const polyBoundaryMesh& allBoundaryMesh,
            const labelList& fromAllTo1Patches,

            const label nInternalFaces,
            const labelList& nFaces,

            labelList& from0ToAllPatches,
            labelList& from1ToAllPatches
        ) const;

        //- Create new zones and add to new mesh.
        static void addZones
        (
            const DynamicList<word>& pointZoneNames,
            const List<DynamicList<label> >& pzPoints,

            const DynamicList<word>& faceZoneNames,
            const List<DynamicList<label> >& fzFaces,
            const List<DynamicList<bool> >& fzFlips,

            const DynamicList<word>& cellZoneNames,
            const List<DynamicList<label> >& czCells,

            polyMesh& mesh
        );

protected:

    // Used by fvMeshAdder

        //- Do all but setting the patches
        List<polyPatch*> addWithoutPatches
        (
            polyMesh& mesh0,
            const polyMesh& mesh1,
            const faceCoupleInfo& coupleInfo
        );

public:

    // Constructors

        //- Construct null
        polyMeshAdder();


    // Member Functions

        const mapAddedPolyMesh& map() const
        {
            return map_;
        }

        //- Add two polyMeshes.
        autoPtr<polyMesh> add
        (
            const IOobject& io,
            const polyMesh& mesh0,
            const polyMesh& mesh1,
            const faceCoupleInfo& coupleInfo
        );

        //- Inplace add mesh to polyMesh.
        void add
        (
            polyMesh& mesh0,
            const polyMesh& mesh1,
            const faceCoupleInfo& coupleInfo
        );
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#ifdef NoRepository
#   include "polyMeshAdderTemplates.C"
#endif

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
