/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description

\*---------------------------------------------------------------------------*/

#include "ComponentMixedTetPointPatchVectorField.H"
#include "constraints.H"
#include "TetPointPatchFieldMapper.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

template
<
    template<class> class PatchField,
    class PointPatch
>
void ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>
::checkFieldSize() const
{
    if
    (
        this->size() != this->patch().size()
     || refValue_.size() != this->patch().size()
     || valueFraction_.size() != this->patch().size()
    )
    {
        FatalErrorIn
        (
            "void ComponentMixedTetPointPatchVectorField::checkField() const"
        )   << "field does not correspond to patch. " << endl
            << "Field size: " << this->size()
            << " value size: " << refValue_.size()
            << " valueFraction size: " << valueFraction_.size()
            << " patch size: " << this->patch().size()
            << abort(FatalError);
    }
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

template
<
    template<class> class PatchField,
    class PointPatch
>
ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>::
ComponentMixedTetPointPatchVectorField
(
    const PointPatch& p,
    const vectorField& iF
)
:
    PatchField<vector>(p, iF),
    refValue_(p.size()),
    valueFraction_(p.size())
{}


template
<
    template<class> class PatchField,
    class PointPatch
>
ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>::
ComponentMixedTetPointPatchVectorField
(
    const PointPatch& p,
    const vectorField& iF,
    const vectorField& v,
    const vectorField& vf
)
:
    PatchField<vector>(p, iF),
    refValue_(v),
    valueFraction_(vf)
{
    checkFieldSize();
}


template
<
    template<class> class PatchField,
    class PointPatch
>
ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>::
ComponentMixedTetPointPatchVectorField
(
    const PointPatch& p,
    const vectorField& iF,
    const dictionary& dict
)
:
    PatchField<vector>(p, iF),
    refValue_("refValue", dict, p.size()),
    valueFraction_("valueFraction", dict, p.size())
{
    this->updateBoundaryField();
}


template
<
    template<class> class PatchField,
    class PointPatch
>
ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>::
ComponentMixedTetPointPatchVectorField
(
    const ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>& ptf,
    const PointPatch& p,
    const vectorField& iF,
    const TetPointPatchFieldMapper& mapper
)
:
    PatchField<vector>(p, iF),
    refValue_(ptf.refValue_, mapper),
    valueFraction_(ptf.valueFraction_, mapper)
{}


template
<
    template<class> class PatchField,
    class PointPatch
>
ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>::
ComponentMixedTetPointPatchVectorField
(
    const ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>& ptf,
    const vectorField& iF
)
:
    PatchField<vector>(ptf, iF),
    refValue_(ptf.refValue_),
    valueFraction_(ptf.valueFraction_)
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

template
<
    template<class> class PatchField,
    class PointPatch
>
void ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>::autoMap
(
    const TetPointPatchFieldMapper& m
)
{
    refValue_.autoMap(m);
    valueFraction_.autoMap(m);
}


// Grab the values using rmap
template
<
    template<class> class PatchField,
    class PointPatch
>
void ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>::rmap
(
    const TetPointPatchField<PatchField, PointPatch, vector>& ptf,
    const labelList& addr
)
{
    const ComponentMixedTetPointPatchVectorField& mptf =
        refCast<const ComponentMixedTetPointPatchVectorField>(ptf);

    refValue_.rmap(mptf.refValue_, addr);
    valueFraction_.rmap(mptf.valueFraction_, addr);
}


// Evaluate patch field
template
<
    template<class> class PatchField,
    class PointPatch
>
void ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>::evaluate()
{
    tmp<vectorField> internalValues = this->patchInternalField();

    // Get internal field to insert values into
    vectorField& iF = const_cast<vectorField&>(this->internalField());

    vectorField values =
        scale(refValue_, valueFraction_)
      + scale(internalValues, vector::one - valueFraction_);

    this->setInInternalField(iF, values);
}


// Set boundary condition to matrix
template
<
    template<class> class PatchField,
    class PointPatch
>
void ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>::
setBoundaryCondition
(
    Map<constraint<vector> > & fix
) const
{
    // get addressing
    const labelList& meshPoints = this->patch().meshPoints();

    // The boundary nod eequation expects a list of valueFractions
    // associated with each component.  In order to enforce
    // consistency, the valueFraction is stored as a vector.

    forAll (meshPoints, pointI)
    {
        const label curPoint = meshPoints[pointI];

        // create a constraint
        constraint<vector> bc =
            constraint<vector>
            (
                curPoint,
                refValue_[pointI],
                cmptMag(valueFraction_[pointI])
            );

        // If pointer is not set, add it, otherwise combine with
        // already existing value
        if (!fix.found(curPoint))
        {
            fix.insert(curPoint, bc);
        }
        else
        {
            fix[curPoint].combine(bc);
        }
    }
}


// Write
template
<
    template<class> class PatchField,
    class PointPatch
>
void ComponentMixedTetPointPatchVectorField<PatchField, PointPatch>::
write(Ostream& os) const
{
    PatchField<vector>::write(os);
    refValue_.writeEntry("refValue", os);
    valueFraction_.writeEntry("valueFraction", os);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
