/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description
    TRI (triangle) file reader. Comes out of e.g. AC3D.
    lines are 9 floats (3 points, each 3 floats) followed by hex colour.
    Is converted into regions: regions numbered from 0 up, each colour is
    region.
    Most of reading/stitching taken from STL reader.
    
\*---------------------------------------------------------------------------*/

#include "triSurface.H"
#include "STLpoint.H"
#include "SLList.H"
#include "IFstream.H"
#include "readHexLabel.H"
#include "stringList.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

bool triSurface::readTRI(const fileName& TRIfileName)
{
    std::ifstream TRIfile(TRIfileName.c_str());

    if (!TRIfile.good())
    {
        FatalErrorIn("triSurface::readTRI(const fileName&)")
            << "Cannot read file " << TRIfileName
            << exit(FatalError);
    }

    SLList<STLpoint> STLpoints;
    SLList<label> STLlabels;
    HashTable<label, string> STLsolidNames;

    // Max region number so far
    label maxRegion = 0;

    while(TRIfile)
    {
        string line = getLine(TRIfile);

        if (line.size() == 0)
        {
            break;
        }

        IStringStream lineStream(line);

        STLpoint p(lineStream);

        if (!lineStream) break;

        STLpoints.append(p);
        STLpoints.append(STLpoint(lineStream));
        STLpoints.append(STLpoint(lineStream));

        // Region/colour in .tri file starts with 0x. Skip.

        char zero;
        lineStream >> zero;

        word rawSolidName(lineStream);

        word solidName("patch" + rawSolidName(1, rawSolidName.size()-1));

        label region  = -1;

        HashTable<label, string>::const_iterator findName =
            STLsolidNames.find(solidName);

        if (findName != STLsolidNames.end())
        {
            region = findName();
        }
        else
        {
            Pout<< "Mapping triangle colour 0" << rawSolidName
                << " to region " << maxRegion << " name " << solidName
                << endl;

            region = maxRegion++;
            STLsolidNames.insert(solidName, region);
        }
        STLlabels.append(region);
    }


    pointField rawPoints(STLpoints.size());

    label i = 0;
    for
    (
        SLList<STLpoint>::iterator iter = STLpoints.begin();
        iter != STLpoints.end();
        ++iter
    )
    {
        rawPoints[i++] = *iter;
    }

    setSize(STLlabels.size());

    label pointI = 0;
    SLList<label>::iterator iter = STLlabels.begin();
    forAll (*this, i)
    {
        operator[](i)[0] = pointI++;
        operator[](i)[1] = pointI++;
        operator[](i)[2] = pointI++;
        operator[](i).region() = *iter;
        ++iter;
    }

    stitchTriangles(rawPoints);

    // Convert solidNames into regionNames
    stringList names(STLsolidNames.toc());

    patches_.setSize(names.size());

    forAll(names, nameI)
    {
        patches_[nameI].name() = names[nameI];
        patches_[nameI].geometricType() = "empty";
    }

    return true;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
